(define *SchemeVersion* "Gambit 2.2")

;;; The following two macros aree borrowed from meroonV3 by Christian Queinnec.
;;; The first macro is a macro definer that allows macros to
;;; remain in the compiled image. The second just defines a macro
;;; local to the sources but normally not visible from the user.

(##define-macro (defmacro call . body)
  `(begin (eval '(##define-macro ,call . ,body))
          (##define-macro ,call . ,body) ) )

(eval '(##define-macro (defmacro call . body)
            `(begin (eval '(##define-macro ,call . ,body))
                    (##define-macro ,call . ,body) ) ))

(##define-macro (defimacro call . body)
  `(##define-macro ,call . ,body) )

 ;; some general new constructs
  
(defmacro (unless x . body) `(if ,x #f (begin ,@body)))
(defmacro (when x . body) `(if ,x (begin ,@body) #f))
(defmacro (for var lower upper . body)
  (let ((loop (gensym)))
   `(let ,loop ((,var ,lower))
      (if (<= ,var ,upper) (begin ,@body (,loop (+ 1 ,var)))))))

(defmacro (while pair . body)
  (let ((label (gensym))
        (var (car pair))
        (lista (cadr pair)))
    `(let ,label ((,var ,lista))
       (cond ((pair? ,var) 
              ,@body
              (,label (cdr ,var)))))))

(defmacro (<- x k)`(vector-ref ,x ,k))
(defmacro (-> x k y)`(vector-set! ,x ,k ,y))

;;=============================================================
;; Defining Blocks and Exiting from them.
(defmacro (block name . code)
   `(call/cc (lambda (,name) ,@code)))
(defmacro (return-from name . arg) 
                `(,name ,(if (null? arg) `#t (car arg))))

;; Syntactic sugar for multiple value return and let
;; (multiple-let ((x y ...)(function with multiple values)) <forms>) 
(defmacro (multiple-let assignment . body)
   (let
     ((list-vars (car assignment))
      (multiple-function (cadr assignment))
      (var (gensym)))
     (define (inner)
       (let((i -1))
         (lambda(x)
           (set! i (+ 1 i))
           (cons x (list `(list-ref ,var ,i))))))
       
     `(let*
        ((,var ,multiple-function)
         ,@(map (inner) list-vars))
        ,@body)))

(defmacro (return-list . lista) `(list ,@lista))

(defmacro (1- x) `(- ,x 1))
(defmacro (1+ x) `(+ ,x 1))
