(require 'queues)
(require 'heaps)
(require 'named-objects)
(require 'events)

;;; MAcros. THis is just the collection of macros defined in $immerEvents
;;; reproduced here for compilation reasons. DO NOT TOUCH.

(##define-macro (+-> x i z) `(##vector-set! ,x ,i ,z))
(##define-macro (<-+ x i) `(##vector-ref ,x ,i))
(##define-macro
  ($object-class? x clase)
  `(= (<-+ ,x 0) (<-+ ,clase ,(find-position 'number class-class))))
(##define-macro ($set-event-dispatcher-clock! x) `(set! *clock* ,x))
(##define-macro ($event-dispatcher-clock) `*clock*)
(##define-macro
  ($event-time elem)
  `(<-+ ,elem ,(find-position 'time event-class)))
(##define-macro
  ($event-priority elem)
  `(<-+ ,elem ,(find-position 'priority event-class)))
(##define-macro
  ($event-message elem)
  `(<-+ ,elem ,(find-position 'message event-class)))
(##define-macro
  ($event-data elem)
  `(<-+ ,elem ,(find-position 'data event-class)))
(##define-macro
  ($set-event-time! elem x)
  `(+-> ,elem ,(find-position 'time event-class) ,x))
(##define-macro
  ($set-event-priority! elem x)
  `(+-> ,elem ,(find-position 'priority event-class) ,x))
(##define-macro
  ($set-event-message! elem x)
  `(+-> ,elem ,(find-position 'message event-class) ,x))
(##define-macro
  ($set-event-data! elem x)
  `(+-> ,elem ,(find-position 'data event-class) ,x))
(##define-macro ($time-round time) 
                `(##inexact->exact (##round (* ,time 1000.))))
(##define-macro ($current-time) `(/ ($event-dispatcher-clock) 1000.))
(##define-macro ($user-event-time aev) `(/ ($event-time ,aev) 1000.))
(##define-macro ($time-from-now t) `(+ ($current-time) ,t))
(##define-macro ($next-event-time) `($user-event-time (peek *lap*)))
(##define-macro
  ($object<?-event x y)
  `(if (##fixnum.< ($event-time ,x) ($event-time ,y))
     #t
     (if (##fixnum.= ($event-time ,x) ($event-time ,y))
       (##fixnum.< ($event-priority ,x) ($event-priority ,y))
       #f)))
(##define-macro
  ($recurrent-event-recurrence-time elem)
  `(<-+ ,elem ,(find-position 'recurrence-time recurrent-event-class)))
(##define-macro
  ($random-recurrent-event-recurrence-time elem)
  `(<-+ ,elem ,(find-position 'recurrence-time random-recurrent-event-class)))
(##define-macro (nevs) `(<-+ *lap* ,(find-position 'n heap-class)))
(##define-macro (lap-size) `(<-+ *lap* ,(find-position 'size heap-class)))
(##define-macro ($peek-event)`(<-+ *EVENT-DATA* 1))


;;;==========================================================================
;;; because of efficiency reasons this is written in terms of *overlord*
;;; using macros to address the fields of the object

;;;------------------------------SIMULATION CONTROL------------------------

(define *overlord* #f)
(define *executing* #f)

;;;======================== PROCESSES =======================================
;;; here everything has been turned to macros for inlining

(define-class coroutine Named-Object
     (( = continuation :initializer (lambda()#f))))

(##define-macro ($coroutine-continuation x)
          `(<-+ ,x ,(find-position 'continuation coroutine-class)))
(##define-macro ($set-coroutine-continuation! x cont)
          `(+-> ,x ,(find-position 'continuation coroutine-class) ,cont))


(##define-macro (SMC:switch aP anotherC Args)
  `(call-with-current-continuation
     (lambda(cont)
       ($set-coroutine-continuation! ,aP cont)
       (set! *executing* ,anotherC)
       (($coroutine-continuation ,anotherC) ,args))))


;;; Processes are frames to allow the interruption of a body of code.
;;; First a few constants that allow the system to know the state of a process

(define iddle 0)
(define waiting-for-event 1)
(define interrupted 2)
(define ready-to-run 3)
(define running 4)

(define-class Process Coroutine 
    ((= status :initializer (lambda()iddle))
     ( = data :initializer (lambda()#f))))

(##define-macro ($set-process-status! x status) 
          `(+-> ,x ,(find-position 'status process-class) ,status)) 
(##define-macro ($process-data x)
          `(<-+ ,x ,(find-position 'data process-class)))
(##define-macro ($set-process-data! x cont)
          `(+-> ,x ,(find-position 'data process-class) ,cont))

;; Processes do not unique identification keys. However, since they could be
;; stored with objects in their data slot, they should inherit their key from
;; whatever is there.

(defmacro ($object-key aP)
     `(object-key ($process-data ,aP)))

(define-method (object<? (aP process)(aP1 process))
     (object<? ($process-data aP)($process-data aP1)))


(define (SMC:Resume aP args) 
  (set! *executing* aP)
  ($set-process-status! aP running)
  (($coroutine-continuation aP) args))

;;; The running process is nowhere. Therefore this forgets about it.

(define (SMC:abandon) (smc:resume *overlord* #f))

;;; Put the process in the ready-list of the dispatcher.

(define (SMC:schedule aP args)
  ($set-process-status! aP ready-to-run)
  ($set-process-data! aP args)
  (put! *READY-LIST* aP))

;;; The main procedure for process switching. Stops the process and calls the
;;; dispatcher. 

(define (SMC:interrupt aP) 
  ($set-process-status! aP interrupted) 
  (smc:switch aP *overlord* #f))

;;; General switching procedure. Puts the process somewhere and calls the 
;;; dispatcher. Special cases are reschedule and wait-duration (see below).
;;; Status is set by assuming that the parking place is not the scheduler!!!

(define (SMC:park aP waiting-place )
  (if (collection? waiting-place) (put! waiting-place aP))
     ($set-process-status! aP interrupted) 
     (smc:switch aP *overlord* #f))
 
;;; Special cases of park:
;;; This puts the process into the ready list and then interrupts it.

(define (smc:reschedule aP args)
  ($set-process-status! aP ready-to-run)
  ($set-process-data! aP args)
  (put! (process-dispatcher-ready-list *overlord*) aP)
  (smc:switch aP *overlord* #f))


(define (smc:Wait-Duration aP time args)
  ($set-process-status! aP waiting-for-event)
  (schedule-resume aP time args) ; to event list 
  (smc:switch aP *overlord* #f))
  
(define (smc:Wait-for-time aP time args)
  ($set-process-status! aP waiting-for-event)
  (schedule-resume-abs aP time args) ; to event list 
  (smc:switch aP *overlord* #f))

;;====================Several MACROS=====================================
;MACROS

;;;syntatic sugar for the definition of the leading character in the play
;;;i.e. Processes 

;;; These macros produces a method or function) than executes asynchronously 
;;; It hides inside a process object that is used to control de activation
;;; of the method.  They should be calles before calling the start-simulation
;;; method. When they are call, they perform initializations and became
;;; suspended until given control by the process dispatcher.
;;; You can control the life of a body of async code by using the functions in
;;; the user interface: abandon, park, schedule, reschedule and interrupt.
;;; You should always use self to refer to a process from inside its definition.
;;; However observe that the call to the method returns the process object,
;;; a very useful feature whenever you need the process, for instance to
;;; interrupt it, or control it otherwise.

(define (process-async-definition definition body type name)
  `(,(if (null? type) 'define (symbol-concatenate 'define- type)) ,definition
     (let ((self (instantiate process 
                              :name (symbol-concatenate ',name '- ',(gensym)))))
       (block init
              (smc:schedule self #f)
              (call/cc (lambda(cont)
                         (set-process-continuation! self cont)
                         (return-from init self)))
              ,@body
              (smc:abandon)))))

(defmacro (define-async-method definition . body)
    (process-async-definition definition body 'method (car definition)))
  
(defmacro (define-async-function definition . body)
    (process-async-definition definition body '() (car definition)))


;;;=====================  PROCESS-DISPATCHERS  =================================

;;; The object that manages other processes. We keep the continuation of the 
;;; dispatcher and resume it. Thus it is a true coroutine. Whenever anybody 
;;; needs the dispatcher, it invokes resume on *overlord*. The dispatcher 
;;; follows continuations. The whole looks like a single very long
;;; dynamically constructed call to dispatch. Eventually, dispatch returns, 
;;; because the ready-list is empty. This returns to the place where the *first*
;;; activation of dispatch was made, all else forgotten in the passing of 
;;; continuations.

(define-class process-dispatcher Process 
     (( = ready-list :initializer (lambda()(instantiate queue)))))


;;; The main generic and method.

(define-generic (smc:dispatch (aK process-dispatcher)))

;;; The method just picks the first element of the raedy-list 
;;; For efficiency reasons this function has all the functionality.

(define-method (smc:dispatch (aK process-dispatcher))
  (let loop ((new (get! *READY-LIST*)))
    (if (not new) #f 
        (begin
         (SMC:switch aK new ($process-data new))
         (loop (get! *READY-LIST*))))))

;;==========================EVENT-DISPATCHERS================================
;;; a process-dispatcher that knowS all about events and time. 
;;; No user methods here
;;;
;;; Global-Continuation is the continuation that stops the simulation.
;;; clock is the simulation clock, time keeping mechanism
;;; event-number: keeps a tally of the number of events and is used to
;;;        identify them uniquely.
;;; Last-event is the last event that happened. It is there for you to check.
;;;     It is useful in, e.g. impatience treatment. See SimmerEvents and test
;;;     program number 11.
;;; LAP-size is the maximum size of the LAP
;;; LAP is the Pending event List (in spanish Lista Acontecimientos Pendientes)


(define-Class Event-Dispatcher process-dispatcher
        (( = global-return :initializer (lambda()#f))
         ( = clock :initializer (lambda()0))
         ( = event-number :initializer (lambda()0))
         ( = last-event :initializer (lambda()#f))
         ( = LAP-size :initializer (lambda()100))
         ( = LAP   :maybe-uninitialized)))

(define-method (initialize! (aEd event-dispatcher))
      (set-event-dispatcher-lap! aEd
             (instantiate heap
               :size (Event-dispatcher-LAP-size aEd))))

;;; Macros to diectionate event-dispatcher fields

(##define-macro ($event-dispatcher-trace)
          `(<-+ *overlord* ,(find-position 'trace  Event-Dispatcher-class)))
(##define-macro ($set-event-dispatcher-last-event! x)
   `(+-> *overlord* ,(find-position 'last-event Event-Dispatcher-class) ,x))

;;; dispatch-events is a function for efficiency  reasons. It is not foreseen
;;; that dispatch events should be speciaalized on the class of event dispatcher
;;; therefore there is not a substantial loss of generality in making this a 
;;; function.



(define (SMC:dispatch-Events) ; very special for *overlord*
  (if (##zero? (nevs)) (error "No pending events at time " ($current-time))
   (let ((x ($peek-event)))
     (if (##fixnum.<= ($Event-priority x) *ignore-value*) ($get-event!)
         (begin
          ($set-event-dispatcher-clock! ($event-time x))
          ($set-event-dispatcher-last-event! x)
          (if ($event-dispatcher-trace) (display-object x))
          (if ($object-class? x recurrent-event-class)
             (begin 
               ($set-event-time! x  
                   (##fixnum.+ *CLOCK* 
                      ($time-round ($recurrent-event-recurrence-time x))))
                  ($replace-event! x))
                ;he eliminado el clone
             (if ($object-class? x random-recurrent-event-class)
               (begin
                ($set-event-time! x 
                   (##fixnum.+ *CLOCK* 
                      ($time-round(($recurrent-event-recurrence-time x)))))
                 ($replace-event! x))
                 ($get-event!)))
          (apply ($event-message x)($event-data x)))))))

           

(define-method (smc:dispatch (aK event-dispatcher))
  (let loop ()
    (call-next-method) ; <=== make-happen returns here recurrently!!!
    (SMC:dispatch-events) 
    (loop)))



 
;;=============================================================================
;;; THE Global Master *overlord* controls everything: time, lifes and destinies
;;=============================================================================
;;; Init simulation initializes the *overlord*. If you provide the Lap size
;;; as the optional argument it is used. else the default size is used. 
;;; The user may also like to initialize *random* by doing 
;;; (set! *random (make-random)) insuring that the random number stream is
;;; the same across runs.
;;; The variables *LAP* y *EVENT-DATA* have been defined in SimmerEvents. 
;;; Here we fill them with values.
;;;

(define (init-simulation . Ev-size)
  (set! *overlord* 
        (if  (null? Ev-size)
          (instantiate event-dispatcher :name "*overlord*")
          (instantiate event-dispatcher :name "*overlord*" 
                       :LAP-size (car Ev-Size))))
  (set! *executing* *overlord*)
  (set! *READY-LIST* (event-dispatcher-ready-list *overlord*))
  (set! *LAP* (event-dispatcher-LAP *overlord*))
  (set! *EVENT-DATA* (heap-a *LAP*))
  (set! *CLOCK* 0))

;;; finally, start the simulation. Store the global continuation to stop
;;; everything dead on its track!.

(define (start-simulation)
  (call/cc
   (lambda(cont)
     (set-Event-Dispatcher-global-return! *overlord* cont)
     (SMC:dispatch *overlord*)))
  'End!!!)

(define (stop) ((event-dispatcher-global-return *overlord*) 'END))



;;;==========================================================================
;;; handy functions that hide *executing*. Useful to control running Processes

(defmacro (Reschedule)`(SMC:Reschedule self #f))

(defmacro (wait-duration time . args)
     (if (null? args) 
       `(smc:wait-duration self ,time #f)
       `(smc:wait-duration self ,time ,@args)))

(defmacro (Wait-for-time time . args)
  (if (null? args)
    `(smc:wait-for-time self ,time #f)
    `(smc:wait-for-time self ,time ,@args)))

;;; and finally the user interface

(define abandon smc:abandon)
(define schedule smc:schedule)
(define interrupt smc:interrupt)
(define park smc:park)


;==========================================================================
