;(##include "meroon.gam")
;(##include "macros.scm")
;(##include "careless.scm")
;(##include "macinst.scm")
;(##include "CompilationStuff.scm")
(require 'diff-equations)
;---------------------------------------------------------------------------
; graphic-differential-object
;---------------------------------------------------------------------------

; Una clase de objetos diferenciales que son capaces de pintar sus trayectorias
; Para ello mantienen un estado grafico compuesto de tantas variables graficas
; de tipo gtime-var, como dimension del estado.

(define-class graphic-differential-object differential-object
   ((= graph-state :maybe-uninitialized)))    ;estado grafico

; crea las variables graficas, incializando su vector. Usa la rutina make&init
; que crea el vector y lo inicializa aplicando una funcion a cada uno de sus 
; elementos. Vector-take realiza un take para todo el vector de estado, que
; inicialmente esta inicializado a cero. Take es un metodo de las stat-vars
; que entrega el valor para que lo guarde la variable. Finalmente declara que
; las variables son del tipo no-step (se pintaran con segemtos de recta).

(define-handy-method (initialize! (aGDO graphic-differential-object))
  (call-next-method)
  (set! graph-state 
        (make&init-vector ns 
             (lambda(x)(instantiate gtime-var :name (qualify-name aGDO x)))))
  aGDO)

; metodo que se llama cuando se quiere entregar el estado al estado grafico. 
; especializando el metodo se consigue que lo que se pinta no sea exactamente
; el estado sino cualquier vector que haya calculado este metodo. Para ver un 
; ejemplo, estudiar la clase var-rate-queue

(define-generic (store-to-draw (aGDO graphic-differential-object)))
(define-method (store-to-draw (aGDO graphic-differential-object))
   (vector-clone (graphic-differential-object-state aGDO)))

; este procedure guarda en el estado grafico de aGDO un vector.
(define (save-graph-data aGDO x0)
  (vector-take (graphic-differential-object-graph-state aGDO) x0
               (graphic-differential-object-t aGDO)))
 
                   
; el post-metodo debe guardar las graficas.
(define-handy-method (output-map (aGDO graphic-differential-object) . args)
   (save-graph-data aGDO (store-to-draw aGDO)))

; finalmente un metodo para pintar todo el estado en la misma grafica.
; Desgraciadamente de esta forma no se obtienen las estadisticas y hay que hacer
; un truco. Todo se arreglaria si el sistema de objetos permitiera
; hace call-next-method de dos niveles mas arriba. La forma que se implementa
; es truculenta y consiste en cambiar el numero de la clase usando el conoci-
; miento de la estructura interna de Meroon, usando project. CUIDADO!!!!. 

(define-handy-method (display-object (aGDO graphic-differential-object) . port)
  (let
    ((lista (vector->list graph-state)))
    (for-each (lambda(x)(display-object (project x 'time-var))) lista)
    (apply display-object-many (graphic-differential-object-name aGDO) lista)))

 
;-------------------------DELTA MAP-------------------------------------------
; el metodo delta-map debe guardar el resultado inmediatamente, ya que no debe
; esperar al proximo output. Lo hace con una discontinuidad.
(define-handy-method (delta-map (aGDO graphic-differential-object) x k)
  (let* ((auxil (store-to-draw aGDO)))
    (call-next-method)
    (let ((s1 (store-to-draw aGDO)))
      (do ((i 0 (+ 1 i)))
          ((= i ns) #t)
        (if (not (= (vector-ref s1 i)(vector-ref auxil i)))
          (take (vector-ref graph-state i)
                (make-discontinuity-point t (vector-ref s1 i) 
                                          (vector-ref auxil i))))))))

 (define-handy-method (dynamic-initialize 
                        (aGDO graphic-differential-object) . args)
   (call-next-method)
   (save-graph-data aGDO (store-to-draw aGDO))
   (do ((i 0 (+ i 1)))
       ((= i ns) #t)
     (reset (<- graph-state i))))

(define-handy-method 
  (start-object-process (aDO graphic-differential-object) t0 . args)
    (create-event t0 dynamic-initialize aDO)
    (create-recurrent-event  t0 25 incrt output-map aDO)
    (create-recurrent-event  t0 100 inct  simulate-step aDO))