(require 'stat-vars)
(require 'simulation)
;
;-------------------------------------------------------------------------
;; A channel is a mail-box where processes leave messages and objects get them. 
;; The mail box is anonymous. Any object getting a message will get the first 
;; mess.in the message queue is there is a message available there. 
;; If there are no messages the requesting process is delayed until a message 
;; becomes available.
;; A process wishing to post a message is never delayed. There is unlimited room
;;for messages in a channel.
;; Not all object can interchange messages. To do so they must have a process
;; (take <object> <message>) that processes them and decides what should be done;; with the message. The methods in channel call take to deliver the message.

;;;========================CHANNELS =====================================
 
(define-class Channel named-object
    ((= queue        :inmutable :maybe-uninitialized)
     (= message-list :inmutable :maybe-uninitialized)))

(define-method (initialize! (aC channel))
      (initialize-field-value!  aC 
         (instantiate stat-queue :name (qualify-name aC "items-waiting"))
         'queue)
      (initialize-field-value!  aC 
         (instantiate stat-queue :name (qualify-name aC "messages-waiting"))
         'message-list)
      aC)

;; take sends a message to a channel
;; give gets a message for the specified object or delays the processs. 
;; When a process gets delayed, the object to receive the message is placed 
;; in the data slot of the suspended process.

(define-method (take (aC Channel) aMessage) 
;non-blocking operation
   (if (empty? (channel-queue aC))
      (put! (channel-message-list aC) aMessage)
     (let
       ((x (get! (channel-queue aC))))
       ( smc:schedule x aMessage))))
                         
(define-method (give (aC Channel) . args) 
;blocking operation!
    (if (empty? (channel-message-list aC))
      (begin
       (if (null? args) #f (set-process-data! *executing* (car args)))
       (smc:park *executing* (channel-queue aC)))
      (get! (channel-message-list aC))))

(define-handy-method (display-object (aC channel) . port)
   (DisplayLine "*****Statistics for Channel " name)
   (display-object queue)
   (display-object message-list))

;;;------------------------a certain kind of message -----------------------
; In simulation this type f message has a long tradition. It carries attributes
; across processes

(define-class Transaction named-object
   ((* attributes :initializer (lambda(x)0))))

(defmacro (make-Transaction) 
             `(instantiate Transaction :attributes-length 10))
;---------------------------------------------------------------------------

                         
