;;SimMerGraphVars.scm
(require 'stat-vars)
(require 'windows)
(require 'time-Functions)

;; Implementing graphical variables. They collect observations, like  normal
;; time variables, but they allow the plotting of the observations in a new
;; window.

;; collects all observations to be plotted. You can select the acceptation
;; of observations by setting the slot accepting. If accepting is false
;; observations are rejected and not collected at all (not even for averaging
;; purposes.


(define-class gtime-var time-var
        ((= data :inmutable :maybe-uninitialized)))

(define-method (initialize! (aGT gtime-var))
    (initialize-field-value! aGT (instantiate time-function) 'data)
    aGT)

;---------------------------------------------------------------------------
; esta clase es la misma que la gtime-var pero se guardan los datos en una
; step-time-function

(define-class stime-var gtime-var ())

(define-method (take (aST stime-var) x)
   (set! x (make-discontinuity-point (point-x x) (point-y x) (var-value aST)))
   (call-next-method))

;---------------------------------------------------------------------------

(define-handy-method (take (agtv gtime-var) x)
  (if accepting  
    (begin
     (if (point? x) (take data x) (Error "Invalid data. Not a point " x))
     (call-next-method))))

(define-generic (GFValue (aGV gtime-var) x))
(define-method (GFValue (aGV gtime-var) x)
  (valor (gtime-var-data aGV) x))
 

;; this is the class of boxes for xy plots. Two labeled axis plus upper
;; and lower lines are drawed. The two points bound the rectangle

(define-class xy-frame graphic-object
    ((= sw-point :maybe-uninitialized)
     (= ne-point :maybe-uninitialized)))

;; This method draws axis and the enclosing box for an x-y plot. The 
;; drawing is done in the plot-area aPa.

(define-handy-method (picture-draw (aF xy-frame) aPa)
  (let
    ((range  (- (point-y ne-point) (point-y sw-point)))
     (ranget (- (point-x ne-point) (point-x sw-point)))
     (frame-ne (make-point (point-x ne-point)(* 1.02 (point-y ne-point)))))
    (if (<= range  0.001) (set! range 0.1))
    (if (<= ranget 0.001) (set! ranget 0.1))
     ;drawing the frame
    (add-go aPa 
        (make-line frame-ne (make-point (point-x frame-ne)(point-y sw-point))))
    (add-go aPa 
        (make-line frame-ne (make-point (point-x sw-point)(point-y frame-ne))))
    (add-go aPa
        (make-line sw-point (make-point (point-x sw-point)(point-y frame-ne))))
    (picture-draw (make-axis 0 sw-point ranget  
                  (exact->inexact (/ ranget 5))
                  (exact->inexact (/ range 100))) aPa)
    (picture-draw (make-axis 1 sw-point range  
                  (exact->inexact (/ range 10))
                  (exact->inexact (/ ranget 100))) aPa)))


;; in addition to the normal display of summary measures, plots the 
;; collected data. Do not interrupt or the system will bomb!

(define-handy-method (display-object (agtv gtime-var) . port)
;;initializations. Reset the world_>window map!!!!
     (call-next-method) 
     (display-object (gtime-var-data agtv) (get-name agtv)))

;; -------------------------------------------------------------------
;; variables to record and display io-curves
;; -------------------------------------------------------------------
(define-class io-var time-var
       ((= in0 :initializer (lambda()0))
        (= out0 :initializer (lambda()0))
        (= input-curve :inmutable :maybe-uninitialized)
        (= output-curve :inmutable :maybe-uninitialized)))

(define-handy-method (initialize! (aio io-var))
  (initialize-field-value! aIO
       (instantiate gtime-var :name (qualify-name aio "Input"))
       'input-Curve)
  (initialize-field-value! aIO
       (instantiate gtime-var :name (qualify-name aio "Output"))
       'Output-Curve)
  (take input-curve (make-point 0 in0))
  (take output-curve (make-point 0 out0))
  aio)

(define-handy-method (increment (aio io-var) . args)
   (call-next-method)
   (take input-curve 
         (make-point (car args) (+ 1 (gtime-var-value input-curve))))) 

(define-handy-method (decrement (aio io-var) . args)
   (call-next-method)
   (take output-curve 
         (make-point (car args) (+ 1 (gtime-var-value output-curve)))))

(define-handy-method (display-object (aiovar io-var) . port)
  (call-next-method) 
  (display-object-many (get-name input-curve)
             input-curve
             output-curve))

;; ---- a step io-var-------------------
(define-class sio-var io-var ())

(define-handy-method (initialize! (aSIO sio-var))
  (initialize-field-value! aSIO
       (instantiate stime-var :name (qualify-name  aSIO "Input"))
       'input-Curve)
  (initialize-field-value! aSIO
       (instantiate stime-var :name (qualify-name  aSIO "Output"))
       'Output-Curve)
  (take input-curve (make-point 0 in0))
  (take output-curve (make-point 0 out0))
  aSio)

;;======================================================================
;; draws many grafvars into the same plotting area. Useful for i-o plots.

(define-method (display-object-list (aGT gtime-var) . lista)
  (apply display-object-list (gtime-var-data aGT) (map gtime-var-data lista)))

; funcion auxiliar.
(define (display-object-many name . lista )
  (set-time-function-header! (gtime-var-data (car lista)) name)
  (apply display-object-list (car lista)(cdr lista)))

 