
(require 'collections)
;
;---------------ORDERED COLLECTIONS -----------------------
; are classes that assume that the stored elements have a method "order<?" 
; that gives a total ordering among the objects.

;The HEAP collection implements a heap and responds to the messages put!, get! 
;expand-heap and peek.
;The heap is stored internally as an array, from indexes 1 to n. key is a  
;closure being used in the comparisons. Several generators are provided 
;covering the common orderings. See below.
;Put! and get! return #f if the operation can not be completed.
;Heap size may be specified by externally creating the array, and can be
;expanded automatically by using the method expand-heap.


(define-class Heap ordered-Collection 
  ((= n    :initializer (lambda() 0))
   (= size :initializer (lambda()200))
   (= a    :maybe-uninitialized)))

(define-handy-method (initialize! (aH Heap))
    (set! a (make-vector size 0))
    aH)

(define-method (put! (unHeap Heap)  args )
  (##declare (standard-bindings) (fixnum j i size heap-n ))
  ;Inserts x into the heap using the ordering specified by key. If no room is
  ;available returns #f.  The type of x must be comparable with previous 
  ;elements in the heap, always using key
  (let
    ((size (- (vector-length (heap-a unHeap)) 1)))
    (define (insert j)
      (if (= j 1) #t
          (let
            ((x (vector-ref (heap-a unHeap) j))
             (i (##truncate (/ j 2))))
            (if
              (object<? x (vector-ref (heap-a unHeap) i))
              (begin
               (vector-set! (heap-a unHeap) j 
                              (vector-ref (heap-a unHeap) i))
               (vector-set! (heap-a unHeap) i x)
               (insert i))))))
    (cond
      ((= (heap-n unHeap) size) #f)
      (else 
       (set-heap-n! unHeap (+ 1 (heap-n unHeap)))
       (vector-set! (heap-a unHeap) (heap-n unHeap) args)
       (insert (heap-n unHeap))
       #t))))


(define-method (peek (unHeap Heap))
"Returns the optimal value without removing it from the heap. If heap is
empty returns #f"
     (if (= (heap-n unHeap) 0)
       #f 
       (vector-ref (heap-a unHeap) 1)))


(define-method (get! (unHeap Heap))
  (##declare (standard-bindings) (fixnum i j m heap-n combina))
  ;Returns the optimal value removing it from the heap. If heap is
  ;empty returns #f
  (let
    ((mayor (vector-ref (heap-a unHeap) 1))
     (x (vector-ref (heap-a unHeap) (heap-n unHeap))))
    (define (combina i)
      (let 
        ((j (* 2 i))
         (m (heap-n unHeap)))
        (cond ((> j m)  i)
              (else 
               (if (and 
                    (< j m)
                    (object<?
                     (vector-ref (heap-a unHeap) (+ 1 j))
                     (vector-ref (heap-a unHeap) j)))
                 (set! j (+ 1 j)))
               (cond
                ((object<?  (vector-ref (heap-a unHeap) j) x)
                 (vector-set! (heap-a unHeap) i 
                                (vector-ref (heap-a unHeap) j))
                 (combina j))
                (else i))))))
    (cond
     ((= (heap-n unHeap) 0) #f)
     (else 
      (set-heap-n! unHeap (-  (heap-n unHeap) 1))
      (vector-set! (heap-a unHeap) (combina 1) x)
      mayor))))
             

(define-method (make-get-elt (aH heap))
    (let
      ((n (heap-n aH))
       (i 1))
      (lambda()
        (let ((j i)) 
          (cond
           ((> i n) #f)
           (else (set! i (+ 1 i))
                 (vector-ref (heap-a aH) j))))))) 


(define-method (expand-size (unH heap) . amm)
       (let*
         ((j (+ (heap-size unH) (if (null? amm) 50 (car amm))))
          (a1 (make-vector j 0)))
         (do ((i 0 (+ 1 i)))
             ((= i (heap-size unH))
              (set-heap-size! unH j)
              (set-heap-a! unH a1))
           (vector-set! a1 i (vector-ref (heap-a unH) i)))))


