
;(require 'channels)
(require 'files)
;
;
;;========================= MAIL BOXES  =====================================
; a mail-box is a place to place messages addressed to an object. They will be 
; delivered whenever the receiving process asks for messages by using
; receive. If a process asks for a message and there are no messages
; addressed to its object, it will be delayed until a message becomes available.
; Matching messages are identified by the key. Messages and objects should
; answer to the object-key method by giving their keys. They are matched to
; deliver the message. 
 
(define-class Mail-box Channel ())

(define-method (initialize! (aMB Mail-Box))
     (initialize-field-value!  aMB 
              (instantiate file :name (qualify-name aMB "processes"))
              'queue)
     (initialize-field-value!  aMB 
              (instantiate file :name (qualify-name aMB "messages"))
              'message-list)
     aMB)

;----------------------------------------------------------------------------
; some auxiliary functions
(define (disp-messages aMB)
  (display "MESSAGES")
  (for-each-elt (mail-box-message-list aMB) (lambda(x)(displayLine x)))
  (newline))

(define (disp-processes aMB)
  (display "PROCESSES")
  (for-each-elt (mail-box-queue aMB)                 
                (lambda(x)(displayLine (process-data x))))
  (newline))

(define (get-my-message aMB anyobj)
  (fread (Mail-Box-Message-list aMB) (object-key anyObj)))

(define (get-my-process aMB aMessage)
  (fread (Mail-Box-Queue aMB) (object-key aMessage)))

(define (remove-message aMB aMessage)
  (remove-elt (Mail-Box-Message-list aMB) (object-key aMessage)))

(define (remove-process aMB aProcess)
  (remove-elt (Mail-Box-Queue aMB) (object-key aProcess)))

;------------------------------------------------------------------------------
;; 
;; send sends a message to a Mail-Box
;; receive gets a message for the specified object or delays it.

(define-method (take (aMB Mail-Box) amessage) 
;non-blocking operation
    (let
      ((x (get-my-process aMB aMessage)))
      (if x
        (begin
          (remove-process aMB x)
          (smc:schedule x aMessage))
        (put! (Mail-Box-Message-list aMB) aMessage))
      aMB))
        
                         
(define-method (give (aMB Mail-Box) . anyobj) 
;blocking operation!
   (let
     ((x (apply get-my-message aMB anyobj)))
     (if x 
       (begin
        (remove-message aMB x) 
        x)
       (begin
        (set-process-data! *executing* (car anyobj))
        (smc:park *executing* (Mail-Box-Queue aMB))))))

(define-handy-method (display-object (aMB Mail-Box) . port)
    (displayLine "*****Mail Box " name)
    (display-object queue)
    (display-object message-list))
;--------------------------------------------------------------------------
(define-class stat-mail-box mail-box ())

(define-method (initialize! (aMB stat-Mail-Box))
     (initialize-field-value!  aMB 
              (instantiate stat-file :name (qualify-name aMB "processes"))
              'queue)
     (initialize-field-value!  aMB 
              (instantiate stat-file :name (qualify-name aMB "messages"))
              'message-list)
     aMB)
