
(require 'files)
;
;;============================SETS=========================================
;are represented as hash tables. The user has the responsibility for insuring 
;that the sizes of the tables are the right ones to perform set operations.
; This is specially important for unions. The system clones the first argument
;therefore the size of the first argument has to be enough to hold the 
;union. However see the method Expand-file in SimmerFiles.scm that can be
;used to expand any file (and therefore a set). 

; A set is simply a file where elements are their own keys. This is insured
; for all objects by the default definiion of object-key, unless object-key
; has been redefined for the class.

(define-class Set File ())

(define-method (put! (aS set)  arg)
  (if (fread aS arg) #f (call-next-method)))

(define-generic (union (as0 set)( as1 set)))
(define-method (union (as0 set)( as1 set))
     (let
       ((x (clone as0))
        (els (make-get-elt as1)))
       (let loop
         ((y (els)))
         (if y (begin
                 (put! x y)
                 (loop (els)))
               x))))
 
(define-generic (intersection (as0 set)(as1 set)))
(define-method (intersection (as0 set)(as1 set))
   (let
     ((x (instantiate set))
      (els (make-get-elt as1)))
     (let loop
       ((y (els)))
       (cond
        ((not y) x)
        ((is-in? as0 y) (put! x y) (loop (els)))
        (else (loop(els)))))))

(define-method (display-object (as Set) . stream)
        (for-each-elt as (lambda(x)(display x)(display #\ ))))

;;  --------------------Stat sets are sets with new methods 
(define-class stat-set set
      ((= number :inmutable :maybe-uninitialized )))

(define-method (initialize! (aS set))
    (init-stats-gatherer aS
         (instantiate time-var :name (qualify-name aS "number-in-set"))
         'number)
    aS)
(define-handy-method (put! (aSS stat-set) arg)
   (let
     ((x (call-next-method)))
     (if x (increment number (current-time)))
     x))

(define-handy-method (remove-elt (aSS stat-set) arg)
  (let
    ((x (call-next-method)))
    (if x (decrement number (current-time)))
    x))



