
(require 'channels)
(require 'knowledge-bases)
;
;
;; a supervisor is an expert in answering questions.
;; It has a channel (mailBox) that receives messages and delivers them.
;; Messsages are questions that should be answered because they are
;; important to the sender. To perform its job, 
;; a supervisor has all parafernalia, required by the toolbox, to be
;; able to reason with rules, plus a data base of frames and links.
;; It also has a list of supervised objects.

(define-class supervisor named-object  
   ((= MailBox :initializer 
               (lambda()(instantiate channel  
                     :queue (instantiate queue)
                     :message-list (instantiate stat-queue :name "messages"))))
    (= info-base :initializer (lambda()'()))
    (= supervised :initializer (lambda()'()))))

(define-handy-method (take (aS supervisor) oneS)
   (set! supervised (cons oneS supervisor)))

(define-generic (answer-me (aS Supervisor) question))
(define-generic (supervisor-dbload (aS Supervisor) .  data))
;------------------------------------------------------

;;A simple knowledge-base supervisor. Has a kb and can answer questions
;;given by patterns. Returns the alist of the first matching fact in 
;;the knowledge base.

(define-class kb-supervisor supervisor ())
(define-handy-method (supervisor-dbload (aKB kb-supervisor) . data)
     (cond
      ((not (pair? data)) #f)
      ((string? (car data)) (loadkb (info-base aKB) data))
      (else
       (set! info-base (apply makekb data)))))

(define-handy-method (answer-me (aS Supervisor) question)
;;here question is a pattern. Returns the alist of the variables.
      (getAList (retrievebyPattern info-base question)))


  
