;;; SimmerExtraControls.scm
;;;
;;;
;;;  =========================  Parallel processing  =========================
;;; THis is a construct to allow parallel processing and synchronization. THis
;;; can be call only from inside an async method. CReates a process, but does 
;;; not stop, t,herefore it can return values to the caller.


(define (process-waitfor-definition definition body type name)
  `(,(if (null? type) 'define (symbol-concatenate 'define- type)) 
     ,definition
     (let ((self (instantiate process 
                              :name (symbol-concatenate ',name '- ',(gensym)))))
       ,@body)))
    

(defmacro (define-waitfor-method definition . body)
    (process-waitfor-definition definition body 'method (car definition)))
  
(defmacro (define-waitfor-function definition . body)
    (process-waitfor-definition definition body '() (car definition)))



;;; This macro is just syntactic sugar to implement a synchronization among
;;; serveral processes. The let body waits for the assignments to finish and
;;; the executes in the local environment. Saves coding!.
;;; Can only be used from inside an async procedure. It is used as follows:
;;;
;;;(let-sync ((x1 (sum1 10))
;;;           (x2 (sum2 11)))
;;;          (displayLine (+ x1 x2)))


(defmacro (let-sync varlist . body)
  (let ((innervars (map (lambda(x)(gensym)) varlist))
        (espera (gensym))
        (procesopadre (gensym))
        (proceso (map (lambda(x)(gensym)) varlist)))
    `(let ((,espera (instantiate gate))
           ,@(Map (lambda(x)`(,x #f)) innervars))
      ,@(map (lambda(x y z)
               `(define-async-function (,z)
                    (set! ,(car x) ,(cadr x))
                    (set! ,y #t)
                    (if (and ,@innervars)(signal ,espera))))
             varlist innervars proceso)
       (define-waitfor-function (,procesopadre)
            ,@(map (lambda(x)`(,x)) proceso)
            (wait-for ,espera)
            ,@body)
       (,procesopadre))))


;;;============================== Processes ===================================
;;;
;;; Defining processes. 
;;;
;;; THis macro creates a process object with the name <name>-obj. 
;;; It also produces a function that when called starts the process. 
;;; The function has the name given in the definition. It allows definition
;;; of the fields in the process like any instantiate form.
;;;
;;; the definition part has the form 
;;;   ((name subtype  <initialization-list>) args)
;;;
;;; subtype is a subclass of process, that can be defined by the user. To start
;;; the process invoque the function with the same name
;;; that is, to run a process you should do: (name)
;;;

(defmacro (define-process definition . body)
   (let*
      ((pro (gensym))
       (namep (caar definition))
       (nameobj (symbol-concatenate namep 'obj))
       (sub-type (cadar definition))
       (rest (cddar definition))
       (arguments (cdr definition)))
      `(let
         ((,pro (instantiate ,sub-type ,@rest :name ',namep)))
         (set-field-value! 
           ,pro 
           (lambda(objetox)
             (lambda ,arguments 
             (if (process-init-code objetox)
               ((process-init-code objetox) ,@arguments))
             (put! (process-dispatcher-ready-list *overlord*) objetox)
             (block init
                    (call/cc (lambda(cont)           
                               (set-process-Continuation! objetox cont)
                               (return-from init))) ; to end of block
                    (let ((self objetox))
                      ,@body)
                    (smc:abandon))) ) 
           'code)
         (set! ,nameobj ,pro)
         (set! ,namep  ((process-code ,pro) ,pro)))))

  
(defmacro (define-init-code nom&args . body)
   `(set-process-init-code! ,(car nom&args) 
        (lambda ,(cdr nom&args) ,@body)))
      
;;; waitfor processes return values to the calling point. They
;;; do not have any init-code since there is no initialization time. They are
;;; called as normal procedures, but allow simulation time to elapse, and
;;; can be suspended or rescheduled.However they always return to the point of
;;; invocation and can return a value.


(defmacro (define-waitfor-process definition . body)
   (let*
      ((pro (gensym))
       (namep (caar definition))
       (nameobj (symbol-concatenate namep 'obj))
       (sub-type (cadar definition))
       (rest (cddar definition))
       (arguments (cdr definition)))
      `(let
         ((,pro (instantiate ,sub-type ,@rest :name ',namep)))
         (set-field-value! 
           ,pro 
           (lambda ,arguments  (let ((self ,nameobj))
                                 ,@body))  
           'code)
         (set! ,nameobj ,pro)
         (set! ,namep (process-code ,pro)))))


;;; Replicate constructs a copy of the given process. Has a syntax similar 
;;; to Meroon's Duplicate. No global symbols are created by this macro, since 
;;; it is intented to be used in the local environment. It returns the function
;;; that is invoked to start the process. No object is returned, therefore 
;;; **you can not** replicate replications of a process. It will give an error 
;;; signalling that the object variable is not defined. 

(defmacro (replicate-process . definition )
   (let* ((pro (gensym))
         (namep (caar definition))
         (nameobj (symbol-concatenate namep 'obj)))
     `(let ((,pro (duplicate ,@(cons (cons nameobj (cdar definition))
                                     (cdr definition)))))
        ((process-code ,pro) ,pro))))

