;;SimmerMenus
(require 'named-objects)
;; This is a very simple implementation of menus. You define a structure
; of menus and submenus and ask install to set-up a menubar for you.
; Changes are suposed to be aplied to a menubar object, and then inserted into
; the menu bar of the mac
; You should redraw the initial menu structure everytime you change it.



(define-class TopMenu named-object 
              (( = initial-number :maybe-uninitialized)
               ( = number   :initializer (lambda()0))
               ( = idnumber :initializer (lambda()0))))

(define *Top-menubar* (instantiate TopMenu :initial-Number 211))

(define-generic (install (aM TopMenu) . where))
(define-generic (deinstall (aM TopMenu) . where))

(define-generic (give-number (aM TopMenu)))

(define-method (give-number (aM TopMenu))
   (let ((x (TopMenu-number aM)))
     (set-TopMenu-number! aM (+ 1 x))
     (+ (TopMenu-Initial-number aM) x)))

;;;--------------------------------------------------------------------------
;;; The structure of the menubar is defined by instantiating meroon object of 
;;; the following kinds

(define-class menubar TopMenu
    ((= menus :initializer (lambda()'()))
     (= to-delete :initializer (lambda()'()))
     ;----------------------------------------
     (= installed :initializer (lambda()#f))
     (= oldmenubar :initializer (lambda()#f))  ; stores here the old menu bar
     (= oldmenuitems :initializer (lambda()'())))) ;and the old control list

(define-method (initialize! (aM menubar))
    (initialize-field-value! aM 211 'initial-number)
    aM)

(define-class menu TopMenu
     ((= string )
      (= menuItems :initializer (lambda()'()))
      ;---------------------------------------
      (= macMenu :initializer (lambda()#f))))

(define-method (initialize! (aM menu))
    (initialize-field-value! aM 1 'initial-number)
    (set-menu-idnumber! aM (give-number *top-menubar*))
    aM)

(define-class MenuItem TopMenu
     ((= string)
      (= action) ))

; In each case you are only required to supply the data above the line. All
; other slots are filled automatically by the initialization process.
; Once you have defined your menus, just call the method install. It will 
; automatically save the old menubar that can be restored using the method 
; reset.
; The action slot should be filled with a thunk, a function of no arguments.
; It will be executed whenever the submenu item is picked.
; if you want more sophisticated things you should use the toolbox to perform
; manipulations. Just remeber that the menus (for Gambit2.2) seem to have the
; following resource id-numbers (for deleting purposes!!!)

   (define appleMenu 128)
   (define fileMenu 129)
   (define editMenu 130)
   (define SearchMenu 134)
   (define CommandMenu 131)
   (define WindowsMenu 132)
   (define SpecialMenu 137)
; 
;just to make sure, check them using the resource editor.
;


;;; --------------Installation method-------------------------------------
;;; saves acopy of the current menu bar and then operates on it, by
;;; deleting and adding menus and items.
;;; MACGAmbit maintains a control list of triples where the action for each
;;; menu item is specified. The list "max#menu-itemshas the form  
;;;              ((macmenu id ) action)
;;; where:
;;;    - MacMenu is the Mac menu where the item is, not the menu object!
;;;    - id is the identification of the item inside the menu a number from 
;;;        zero onwards
;;;    - action is a thunk that will be executed if the item is selected.
;;; So what isntall does is construct the menu items, put them into the menus
;;; put the menus into the menubar and add to the control list.

;;;
;;; Managing the MacGambit control list--------------------------------------
;;; function that adds a triple  ((macmenu id) action) to the control list

(define (menu-key amenu amenuitem) 
  (cons (menu-macmenu amenu) (menuitem-idnumber amenuitem)))

(define (add-control amenu amenuitem)
  (set! mac#menu-items 
        (cons (cons (menu-key amenu amenuitem) 
                    (menuitem-action amenuitem)) mac#menu-items)))


;;; Managing a menu object. Use nconc because we want to add at the end of 
;;; the item list

(define (add-item-to-menu amenu aMenuItem)
  (set-menu-menuitems! aMenu (nconc (menu-menuItems amenu) (list amenuItem))))

(define (remove-item-from-menu amenu aMenuItem)
  (set-menu-menuitems! aMenu (remove aMenuItem (menu-menuItems aMenu))))


;;; The Mac menu is what is physically used by the OS to dispatch menuitems
;;;
;;; Managing the mac menu in a Menu-------------------------------------------
;;;
(define (put-in-mac-menu aMenu aMenuitem)
  (mac#insertmenuitem (menu-macMenu amenu) 
                      (menuItem-string aMenuitem) 
                      (menuItem-idnumber aMenuItem)))

(define (delete-from-mac-menu aMenu aMenuitem)
  (mac#deletemenuitem (Menu-macmenu amenu) 
                      (menuItem-idnumber aMenuItem)))
;;;----------------------------------------------------------------------------
;;;

;;; The instalation methods

(define-method (install (aMI menuitem) . aM)
   (let* ((amenu (car aM))
          (n (give-number amenu)))
     (if n
       (begin
        (set-menuitem-idnumber! aMI n)
        (put-in-mac-menu aMenu aMI)
        (add-control amenu aMI))))
   #t)

 
(define-method (install (aM menu) . aMB)
   (set-menu-number! aM 0)
   (let* ((n (menu-idnumber aM)))
     (set-menu-macMenu! aM (mac#newmenu n (menu-string aM)))
     (for-each (lambda(x)(install x aM)) (menu-Menuitems aM))          
     ;insert menu at end of menu bar
     (mac#insertmenu (menu-macMenu aM) 0))
   #t) 

(define (initialize-menus-to aMB)
  (set-menubar-oldmenubar! aMB (mac#getmenubar))
  (set-menubar-oldmenuitems! aMB mac#menu-items)
  (set-menubar-installed! aMB #t)
  (install aMB))
  
(define-method (install (aMB menubar) . where) 
  (if (menubar-installed aMB)
    (begin
     (set-menubar-number! aMB 0)
     (set! mac#menu-items (menubar-oldmenuitems aMB))
     (mac#setmenuBar (menubar-oldmenubar aMB))
     (for-each (lambda(x)(mac#deletemenu (menu-macmenu x)))
               (menubar-to-delete aMB))
     (for-each (lambda(x)(install x aMB)) (menubar-menus aMB))
     (mac#drawmenubar)))
     #t)

;; this is a physical and brutal reset
(define-method (reset (aMB menubar))
  (mac#clearmenubar)
  (set! mac#menu-items (menubar-oldmenuitems aMB))
  (mac#setmenuBar (menubar-oldmenubar aMB))
  (mac#drawMenuBar)
  (set-Menubar-installed! aMB #f))

;---------------------------------------------------------------------------
;; This is a simple exemple. Try it
;; for a more complicated one look at the file SimmerEnvironment.scm in 
;; folder Suplement
;---------------------------------------------------------------------------
;(define anewbar
;  (instantiate menubar
;     :to-delete 0
;     :menus  
;      (list
;        (instantiate menu 
;           :string "First"
;           :menuitems 
;             (list
;                (instantiate menuitem
;                          :string "First-a"
;                          :action (lambda()(displayLine "First-a")))
;               (instantiate menuitem
;                          :string "First-b"
;                          :action (lambda()(displayLine "First-b")))
;               (instantiate menuitem
;                          :string "First-c"
;                          :action (lambda()(displayLine "First-c"))))))
;       (instantiate menu 
;           :string "Segundo"
;           :menuitems 
;             (list
;                (instantiate menuitem
;                          :string "Second-a"
;                          :action (lambda()(displayLine "Second-a")))
;               (instantiate menuitem
;                          :string "Second-b"
;                          :action (lambda()(displayLine "Second-b")))))))



;(install anewbar)                                    
;(reset anewbar)    
