;; SimmerRegisteredClasses
(require 'files)
;; This is to register instances of objects. Its done by defining the object
;; class as a registered class. This creates a diccionary of objects and allows 
;; registration and setup.
;==============================================================================
;;             Machinery to create diccionaries 

;;---------------REGISTERED CLASS-------------------------------------------
;; A registered class is a class that has a dictionary associated, where 
;; instances of the class (or its descendents) are indexed by their
;; name

(define-class registered-class meroonv2-class 
     ((= dictionary :initializer (lambda()(instantiate file :size 100)))))

;; this function will register an object in class clase using its key. This
;; is udeful in initialization methods to keep a roll of the defined objects.

(define (register-in-class clase objeto)
  (if (is-a? objeto clase) (put! (class-dic clase) objeto)
      (error "Objet does not belong to class " (named-object-name objeto) " "
             (class-name clase))))

; find the first registered superclass for that object.
(define (find-registered-superclass objeto)
  (let loop ((x (object->class objeto)))
    (if (registered-class? x) x
        (let ((y (class-super-number x)))
          (if y (loop (->class y))
              #f)))))

;this registers the object in the first registered superclass
(define (register objeto)
  (register-in-class (find-registered-superclass objeto) objeto))

;----------------------------------
; An instantiate macro that registers the resulting object
(defmacro (r-instantiate . definition)
   `(let ((x (instantiate ,@definition)))
      (register x)
      x))

; functions to acces a diccionary and an keyed element in it
; the next function returns de diccionary associated witha registered class.
(define (class-dic x)(registered-class-dictionary x))

;This functions looks up an object key into the class dictionary of class aRC
; and returns the object
(define (fetch  aRC name)
    (fread (registered-class-dictionary aRC) name))

; This generic function reinitializes(and clears) the class dictionary.
(define-generic (reset-class (aC class)))
(define-method (reset-class (aRC registered-class))
   (set-registered-class-dictionary! aRC (instantiate file :size 100))
   #t)

;;============================================================================
;; Some USEFUL MACROS AND FUNCTIONS

; a macro to reset several classes at the same time, just by giving their names
(defmacro (reset-classes . lista-nombres)
  `(begin 
   ,@(map 
      (lambda(x) `(reset-class ,(symbol-concatenate x '-class)))
      lista-nombres)))

;; perform a function on all elements of a class
(define (do-function-on aclass f)
  (for-each-elt (class-dic aclass) f))

; a macro to start a process on all elementes of several classes
(defmacro (start-method-on-classes process . classes)
  (let ((uno (gensym)))
    `(begin
      ,@(map 
          (lambda(y) 
            `(for-each-elt 
               (class-dic ,(symbol-concatenate y '-class))
               (lambda(,uno)(,process ,uno))))
          classes))))
  
; a macro to perform start-object-process on all classes
(defmacro (start-object-processes time . classes)
   (let ((uno (gensym)))
     `(if (number? ,time)
        (begin
         ,@(map 
             (lambda(y) 
               `(for-each-elt 
                  (class-dic ,(symbol-concatenate y '-class))
                  (lambda(,uno)(start-object-process ,uno ,time))))
             classes))
        (error "Tiempo invalido arrancando procesos de " classes))))
  

; a function to find a key in any of several dictionaries
(define (key->object akey . dictionaries)
  (let loop ((x dictionaries))
    (if (pair? x)
      (let ((y (fread (car x) akey)))
        (if  y y (loop (cdr x))))
      #f)))

;invierte en slotb, la relacion entre las clases clasa y classb, dada por el
; slota de la clase classa.

(define (invert-relationship classa fielda classb fieldb)
  (for-each-elt (registered-class-dictionary classa)
      (lambda (xx) 
        (map 
          (lambda(x)
            (let ((y (fetch classb (object-key x))))
              (if y 
                (set-field-value! y (cons xx (field-value y fieldb)) fieldb)
                (error "Inexistent object :" (object-key x) 
                       " in class "  (class-name classb)))))
          (field-value xx fielda)))))
              
