;;SimmerWindows.scm
(require 'named-objects)
(require 'menus)
(require 'graphics)
(require 'files)
;;--------------------------------------------------------------------------
;; this is an adaptation of one of the exemple programs that comes with
;; MacGambit. Opens a window and creates a manager for the window. The
;; manager has simple facilities to handle some events.
;; To use, just create  the windows and call manage-window on each one of them.
;; You have to provide functions to update, close, etc.
;;;; There is also an elementary menu facility, that collects all opened windows
;; in a "Results" menu.
;; To use it you should initialize the facility by using init-results-menu. From
;; then on everything will be automatic. In case of a failure (an error), 
;; for instance, you should restore the in initial state by using 
;; reset-results-menu 


;;; windows are kept together with their menu entries in the slot open-windows 
;;; from the unique window-system object
;;; this is the object (record) for a typical entry in the list. When a window
;;; is open ths record is created, added to the list and its menu item added
;;; to the menu.

(define-class window-record object
   (theWindow theMenuitem))

(define-method (object-key (aWR window-record))
   (window-record-theWindow aWR))

(define-class window-system menubar
  ((= window-num :initializer (lambda()1))
   (= base-position :initializer (lambda()(list 50 50 270 350)))
   (= open-windows :initializer (lambda()(instantiate file)))))

(define (results-menu sistema)
  (car (window-system-menus sistema)))

(define-generic (close (aWS window-system) w))

(define-method (close (sistema Window-system) w)
  (let ((x (fread (window-system-open-windows sistema) w)))
    (if x 
      (let ((aMenuItem (window-record-theMenuItem x)))
        (remove-elt (window-system-open-windows sistema) w)
        (remove-item-from-menu (results-menu sistema) aMenuItem )
        (mac#window-unbind w)
        (mac#disposewindow w)
        (install sistema)
        (mac#drawmenubar))
      (error "Attempting to close an unopened window " (mac#getwname w)))))
      
(define (close-all-result-windows)
  (for-each-elt (window-system-open-windows *results*)
                (lambda(y) (close *results* (window-record-theWindow y)))))


;;; special functions for *results*
(define *results* 
  (instantiate window-system
      :menus 
      (list 
        (instantiate menu
                :string "Results"
                :menuItems
                      (List 
                        (instantiate MenuItem
                           :string "Clean"
                           :action (lambda()(close-all-result-windows)))
                        (instantiate MenuItem
                            :string "____________"
                            :action (lambda()#f)))))))

(define (window-number)
  (let ((n (window-system-window-num *results*)))
    (set-window-system-window-num! *results* 
                                   (+ 1 (window-system-window-num *results*)))
    n))
                                   
;; staircases the windows as they are created.
(define (window-rectangle i)
  (apply mac#rect 
         (map (lambda(x)(+ x (* (modulo i 20) 10))) 
              (window-system-base-position *results*))))


;; physical making of a window.
(define (make-window rect name)
  (let
    ((w (mac#newwindow rect name #t 4 -1 #t)))
    (if (= w 0) (error "Window could not be created (out of memory?)"))
    w))

;; this method belogs to pantalla but is here to pull all windowing things here

(define-generic(manage-window (pantalla plot-area))) 
(define-method (manage-window (pantalla plot-area))
  (let
    ((w (plot-area-port pantalla)))
    ; These procedures handle the events generated by the Macintosh

    (define (mousedown pt modifiers) (action "mousedown"  #f modifiers pt))
    (define (mouseup pt modifiers)   (action "mouseup"    #f modifiers pt))
    (define (keydown ch modifiers)   (action "keydown"    ch modifiers #f))
    (define (keyup ch modifiers)     (action "keyup"      ch modifiers #f))
    (define (autokey ch modifiers)   (action "autokey"    ch modifiers #f))
    (define (update)                 (update-graph))
    (define (activate)               (action "activate"   #f #f        #f))
    (define (deactivate)             (action "deactivate" #f #f        #f))
    (define (goaway)                 (bye))
    (define (zoomin)                 (action "zoomin"     #f #f        #f))
    (define (zoomout)                (action "zoomout"    #f #f        #f))
    (define (grow)                   (action "grow"       #f #f        #f))

    ; Utilities for above
    (define (action event-name char modifiers pt) #t)
    (define (bye) 
      (if (window-system-installed *results*) 
        (close *results* w)
        (begin
          (mac#window-unbind w)
          (mac#disposewindow w))))
      
    
    (define (update-graph)
      (mac#eraserect w (mac#rect -32000 -32000 32000 32000))
      (draw pantalla))

    ; Create the event handler

    (define (handler event)
      (let ((what (mac#event-what event)))
        (case what
          ((1) ; mousedown
           (mousedown
             (mac#event-where event)
             (mac#event-modifiers event)))
          ((2) ; mouseup
           (mouseup
             (mac#event-where event)
             (mac#event-modifiers event)))
          ((3) ; keydown
           (keydown
             (integer->char (##logand (mac#event-message event) 255))
             (mac#event-modifiers event)))
          ((4) ; keyup
           (keyup
             (integer->char (##logand (mac#event-message event) 255))
             (mac#event-modifiers event)))
          ((5) ; autokey
           (autokey
             (integer->char (##logand (mac#event-message event) 255))
             (mac#event-modifiers event)))
          ((6) ; update
           (update))
          ((8) ; activate/deactivate
           (if (odd? (mac#event-modifiers event))
             (activate)
             (deactivate)))
          ((256) ; goaway
           (goaway))
          ((257) ; zoomin
           (zoomin))
          ((258) ; zoomout
           (zoomout))
          ((259) ; grow
           (grow)))))

    ; If window was created, announce its presence to the window manager

    (if (= w 0)
      (error "Window could not be created (out of memory?)")
      (begin
        (mac#window-bind w handler)
        (if (window-system-installed *results*) (take *results* w))
        (update-graph)
        handler))))


(define (create-item-and-put-in-menu sistema nombre funcion)
  (let ((item (instantiate MenuItem :string nombre :action funcion)))
     ;; instala la ventana en el menu results
    (add-item-to-menu (results-menu *results*) item)
    item))

    
(define-method (Take (sistema window-system) ventana)
  (let ((thefuncion (lambda() 
                      (mac#showwindow ventana) 
                      (mac#selectwindow ventana)))
        (nombre (mac#getwtitle ventana)))
    ; If needed a window is given a unique name derived from the pointer! 
    (if (string=? nombre "") 
      (set! nombre (string-append "Window-" (number->string ventana))))
    ;;; creates a menuitem for the window in the menu in sistema
    (let ((anItem (create-item-and-put-in-menu sistema nombre thefuncion)))
    ;;; record the window and menu in the open-windows list
      (put! (window-system-open-windows sistema) 
            (make-window-record ventana anItem)))
    (install *results*))) 
  
                             
(define (remove-results-menu) (reset *results*))

(define (reset-drawing ventana) (mac#window-reset ventana))

(define (start-result-windows)(initialize-menus-to *results*))