;;; Uncomment the following line if you are not using a compiled version
;(require 'stat-vars)
;(require 'mail-boxes)

; There are three objects. One of them is the source of messages to the channel 
; post-office. The other two objects  act as sinks for the messsages, ask for 
; messages and go away with their slots filled. . There are two methods, one 
; sending and another receiving. We need a method take, that puts the message 
; into the asking object

; The structure of the program is
;      1) Define the class of objects.
;      2) define the method take plugging the messages into the object.
;      3) Define the asynchronous methods
;      4) The message is a predefined type, class transaction with key given 
;         by name
;      5) Instantiate everything and go.
;
; The neat thing is that the use of async methods allows many instances of 
; the receiving objects to populate the system, and share the behavior specifyed
; by the method.

(define-class messenger named-object
      ((= wait :initializer (lambda()5))
       (= numero :initializer (lambda()0))
       (= message :initializer (lambda()#f))))


(define-generic (send-message (aR messenger)))
(define-generic (get-message (aR messenger)))

(define-async-method (send-message (aR messenger)) 
  (let loop ()
    (for i 1 2
         (let((m1 (instantiate transaction :name 'tran :attributes-length 10)))
           (set-transaction-attributes! m1 0 (current-time))
           (set-transaction-attributes! m1 1 (messenger-numero aR))
           (set-messenger-numero! aR (+ 1 (messenger-numero aR)))
           (if (< (*random*) 0.7) 
             (set-transaction-name! m1 'p2)
             (set-transaction-name! m1 'p3))
           (displayLine "sending: " m1)
           (take post-office m1)))
    (wait-duration (messenger-wait aR))
    (loop)))

(define-async-method (get-message (aR messenger)) 
  (let loop ()
    (displayLine "asked " (messenger-name aR))
    (set-messenger-message! aR (give post-office aR))
    (displayLine "receiving " (messenger-name aR) ":" (messenger-message aR))
    (wait-duration (messenger-wait aR))
    (loop)))


(begin
 (init-simulation)
 (set! *random* (make-random))
 (define post-office (instantiate stat-mail-box :name 'Central))
 (define p1 (instantiate messenger :name 'p1 :wait 5))
 (define p2 (instantiate messenger :name 'p2 :wait 3))
 (define p3 (instantiate messenger :name 'p3 :wait 2))
 (send-message p1)
 (get-message p2)
 (get-message p3)
 (create-event 6 stop)
 (trace-off *overlord*)
 (start-simulation)
 (display-object post-office))