;;; Uncomment the following 2 lines if you are not using a compiled version
;(require 'channels)
;(require 'distributions)
;;;--------------------------- Simulating a FIFO queue ---------------------
;;;
;;; This is an implementation of queues using a pool of servers, defined
;;; as a channel this is probably the simplest of queue implementations.
;;; customers are generated and put! into a queue, the queue "system", that is
;;; used to get statistics. Inmediately they request a message from the channel
;;; (the pool) of resources. The message is a server that is returned once the
;;; service has completed. 
;;;
;;;-------------------------------------------------------------------------
(define-class service-center channel 
      ((= system :maybe-uninitialized)
       (= dist-arrival :initializer (lambda()#f))))
  
(define-method (initialize! (aSC service-center))
    (call-next-method)
    (set-service-center-system! aSC
      (instantiate stat-queue :name (qualify-name aSC 'in-system)))
    aSC)


(define-class server-item object 
      (( = dist-service :initializer (lambda()#f))))
(define-class customer object 
      ((= server-assigned :initializer (lambda()#f))))

(define-generic (service (aC customer) aservice-center))
(define-generic (main (aSc service-center)))


;;  the life story of an item
  (define-async-method (service (aC customer) aservice-center)
     (put! (service-center-system aservice-center) aC)    ;store the customer
     (set-customer-server-assigned! aC (give aservice-center)) ;get server
     (wait-duration 
       (sample (server-item-dist-service  
                 (customer-server-assigned aC))))   ;now wait
     (take aservice-center (customer-server-assigned aC)) ;give back the server
     (get! (service-center-system aservice-center)))  ;remove the customer 

;; the main program that generates items
  (define-async-method (main (aSC service-center))
      (let loop ()
        (service (instantiate customer) aSC) 
        (wait-duration (sample (service-center-dist-arrival aSC)))
        (loop)))

(let
  ((aService (instantiate service-center
                          :name 'Service-point
                          :dist-arrival (make-exponential 10)))
   (t0 (runtime)))  
  (init-simulation)
  (set! *random* (make-random))
  (show-generic-untrace 'take 'give)
  (trace-off *overlord*)
  (create-event 1000 stop)
  (for i 1 1 (take aService (instantiate server-item  
                                         :dist-service (make-exponential 8))))
  (main aService)
  (start-simulation)
  (displayLine "Elapsed Time " (formatf (- (runtime) t0) 6 2) " seconds")
  (newline)
  (display-object aService)
  (newline)
  (display-object (service-center-system aService)))

