;; TESTING FACILITIES
;----------------------------------------------------------------------------
;;; Uncomment the following line if you are not using a compiled version
; (require 'monitors)
;;; This implements a queue with n servers using facilities. The server is 
;;; modeles as a facility, therefore all servers are identical. We redefine 
;;; the entry queue to keep statistics on wait.

(define distribution (make-normal 12 2))
(define service-dist 4)
(define (numer) 
  (let ((n 0))
    (lambda()(let ((x n))(set! n (+ 1 n)) x))))

;;;---------------------------------------------------------------------------

;;; In this very simple model we only need two objects, the queue and the 
;;; arriving customers.

(define-class product named-object
    ((= n :maybe-uninitialized )
     (= processing-time :initializer (lambda() (sample distribution)))))

;;;---------------------------------------------------------------------------
;;; And two methods, the method to process each item and the method to 
;;; generate arrivals.
;;; The method to process an item

(define-generic (working (aP product) machine))
(define-async-method (working (aP product) machine)
    (if (facility-trace machine)
      (displayLine (product-n aP) " trato de entrar " (current-time)))
    (enter machine)
    (if (facility-trace machine)
      (displayLine (product-n aP) " entro " (current-time)))
    (wait-duration (product-processing-time aP))
    (if (facility-trace machine)
      (displayLine (product-n aP) " salgo " (current-time)))
    (leave machine))

;;; The arrival process

(define-async-function (product-arrival machine)
   (let loop ((x (instantiate product :n (numer)))) 
     (working x machine)
     (wait-duration (sample service-dist))
     (loop (instantiate product :n (numer)))))

;;;---------------------------------------------------------------------------
;;;  Now simulate 
;;; if you want to see what is going on, do (trace-on aMachine)
;;;---------------------------------------------------------------------------
(define aMachine  
  (instantiate facility 
     :name 'maquina
     :max-number 3
     :parking-place 
     (instantiate stat-queue :name 'esperando-entrar)))
(init-simulation)
(set! *random* (make-random))
(create-event 100 stop)
(product-arrival aMachine)
(start-simulation)
(display-object (facility-number amachine))
(display-object (facility-parking-place amachine))
     