;;; Uncomment the following 3 lines if you are not using a compiled version
;(require 'channels)
;(require 'distributions)
;(require 'graph-vars)

; This is an implementation of queues using a pool of servers, defined
; as a channel. We provide elaborate statistics and graphics.
; This is the simplest of queue implementations for n(non-identical)-servers.
; customers are generated and put! into a queue, the queue "system", that is
; used to get statistics. Inmediately they request a message from the channel
; (the pool) of resources. The message is a server that is returned once the
; service has completed. 

(define-class service-center channel 
      ((= system :initializer 
          (lambda()(instantiate stat-queue :name 'system)))
       ( = dist-arrival :initializer (lambda()#f))))

(define-method (initialize! (aSC service-center))
   (call-next-method)
   (set-stat-queue-number! (service-center-queue aSC)
          (instantiate stime-var :name (qualify-name aSC 'system-stats)))
   (set-stat-queue-number! (service-center-message-list aSC) 
          (instantiate stime-var :name (qualify-name aSC 'server-stats)))
   aSC)
   
(define-generic (main (aSc service-center)))

(define-class server-item object 
      (( = dist-service :initializer (lambda()#f))))

(define-class customer object 
              ((= server-assigned :initializer (lambda()#f))))

(define-method (take (aC customer) aserver)
   (set-customer-server-assigned! aC aserver))

(define-generic (service (aC customer) aservice-center))


;;  the life story of an item
(define-async-method (service (aC customer) aservice-center)
   (put! (service-center-system aservice-center) aC)    ;store the customer 
   (set-customer-server-assigned! aC (give aservice-center))   ;get server
   (wait-duration 
     (sample (server-item-dist-service  
               (customer-server-assigned aC))))   ;now wait
   (take aservice-center (customer-server-assigned aC)) ;give back the server
   (get! (service-center-system aservice-center)))  ;remove the customer 

;; the main program that generates items
(define-async-method (main (aSC service-center))
   (let loop ()
     (service (instantiate customer) aSC) 
     (wait-duration (sample (service-center-dist-arrival aSC)))
     (loop)))

;;;
;;; define the objects
;;;
(define unservice (instantiate service-center :dist-arrival 
                               (make-exponential 10)
                               :name 'Service-Center)) 
(define n 1)   ;number of servers
;;; 
;;; initialization
;;;
(init-simulation)
(set! *random* (make-random))
(trace-off *overlord*)
(create-event 1000 stop)
(for i 1 n (take unservice (instantiate server-item  
                                        :dist-service (make-exponential 8))))
(main unservice)
;;;
;;; Perform simulation
;;;
(let ((t0 (runtime)))
  (start-simulation)
  (displayLine "Runing Time " (- (runtime) t0)))
;;;
;;; Now display statistics and graphs
;;;
(newline)
(display-object unservice)
(newline)
(display-object (service-center-system unservice))

