;(require 'monitors)
;(require 'distributions)
;(require 'graph-vars)
;--------------------------------------------------------------------
;; This program is mainly documented in spanish. It is offered as an exemple
;; of a (simplified) real world simulation. It is actually based on a 
;; case I heve been workin on.

;; first subclass the simulation classes stat-queue and facility to
;; allow for graphics

(define-class gstat-queue stat-queue())
(define-method (initialize! (agsQ gstat-queue))
      (call-next-method)
      (set-field-value! agsQ 
           (instantiate stime-var :name (qualify-name agSQ  'stats))
           'number)
      agsQ)

(define-class stat-facility facility ())
(define-method (initialize! (asC stat-facility))
    (call-next-method)
    (set-field-value! aSC 
        (instantiate gstat-queue :name (qualify-name aSC 'parking-place))
        'parking-place)
    aSC)

;---------------------------------------------------------------------
;; define the two simulation classes.

(define-class paciente named-object
         ((= clase :initializer (lambda()0))
          (= lista-departamentos :initializer(lambda()'()))))

(define-class departamento stat-facility
      ((= service-distribution :maybe-uninitialized)))

(define-class hospital named-object
         ((= num-departamentos)
          (= num-clases-clientes)
          (= selector-clientes :maybe-uninitialized)
          (= clases :maybe-uninitialized)
          (= departamentos :maybe-uninitialized)
          (= total-clientes :initializer (lambda()0))
          (= total-servidos :initializer (lambda()0))))

;; crea una lista de distribuciones para generar el numero de visitas a
;; cada departamento de un cliente.

(define (make-llegadas-de-esta-clase lista-visitas coefvar)
  (map 
    (lambda(x)(if (zero? x) 0 (make-log-normal x (* coefvar x))))
   lista-visitas))


(define-method (initialize! (aH Hospital))
    (set-hospital-total-clientes! aH 0)
    (set-Hospital-clases! aH 
        (make-vector (hospital-num-clases-clientes aH)))
    (set-Hospital-selector-clientes! aH 
        (make-discrete '(0 1 2 3) '(0.1 0.3 0.4 0.2 )))   
    (vector-set! (hospital-clases aH) 0 
        (make-llegadas-de-esta-clase '(1 2 0.5 0 0.04) 0.5))
    (vector-set! (hospital-clases aH) 1 
        (make-llegadas-de-esta-clase '(2 1 0.1 0 0.1) 0.5))
    (vector-set! (hospital-clases aH) 2 
        (make-llegadas-de-esta-clase '(0 0.5 0 2 0.04) 0.5))
    (vector-set! (hospital-clases aH) 3 
        (make-llegadas-de-esta-clase '(1 0 0.5 1 0.2) 0.5))
    (set-Hospital-departamentos! aH 
        (make-vector (Hospital-num-departamentos aH)))
    aH)

(define-generic (IncNumTotal (aH Hospital)))
(define-handy-method (IncNumTotal (aH Hospital))
     (set! total-clientes (+ 1 total-clientes)))

(define-generic (IncNumServidos (aH Hospital)))
(define-handy-method (IncNumServidos (aH Hospital))
     (set! total-servidos (+ 1 total-servidos)))

(define-generic (Nombre-Prox-Cliente (aH Hospital)))
(define-handy-method (Nombre-Prox-Cliente (aH Hospital))
     (string-append "Cliente "
                    (number->string Total-Clientes)))

(define-method (display-object (aH Hospital) . stream)
    (newline)
    (displayLine "*****DEPARTAMENTOS****")
    (map (lambda(x)(display-object x)) 
         (vector->list (Hospital-departamentos aH))))

(define-generic (genera-cliente (aH Hospital)))
(define-method (genera-cliente (aH Hospital))
    ;; genera una muestra del numero de visitas
    (define (muestra-de-la-clase w) 
      (map (lambda(x)(inexact->exact(round (sample x)))) w))
    
    ;; convierte una lista de numero de visitas a departamentos 
    ;;en una lista de departamentos a visitar, por orden de visita.
    (define (make-proximo lista)
      (define (suma list)
        (if (pair? list) (+ (car list)(suma (cdr list))) 0))
      (let* ((n (suma lista))
             (in-vect (list->vector lista))
             (out-vect (make-vector n))
             (m (vector-length in-vect)))
        (do ((j 0 (modulo (+ 1 j) m))
             (i 0 i))
            ((= i n) (vector->list out-vect))
          (if (> (vector-ref in-vect j) 0)
            (begin
             (vector-set! in-vect j (- (vector-ref in-vect j) 1))
             (vector-set! out-vect i j)
             (set! i (+ 1 i)))))))
    (let
      ((cliente (instantiate paciente 
           :clase (sample (Hospital-selector-clientes aH))
           :name (Nombre-Prox-Cliente aH))))
      (IncNumTotal aH)
      (set-paciente-lista-departamentos! cliente 
           (make-proximo 
            (muestra-de-la-clase 
             (vector-ref (Hospital-clases aH) (paciente-clase cliente)))))
      cliente))


;; Definimos el metodo de recorrido de un paciente

(define-generic (procesa (aP paciente) aH))

(define-async-method (procesa (aP paciente) aH)
   (let ((First-Dept (lambda(lista)
                       (vector-ref (hospital-departamentos aH) (car lista)))))
     (let* ((x (genera-cliente aH))
            (n (length (paciente-lista-departamentos x))))
       (let loop ((lista (paciente-lista-departamentos x)))
         (if (pair? lista)
           (begin
            (IncNumServidos aH)
            (let
              ((dept (First-Dept lista)))
              (enter dept)
              (wait-duration (sample (departamento-service-distribution Dept)))
              (leave dept))
            (loop (cdr lista))))))))


;;Simula es el proceso que lanza los clientes al hospital.
(define-async-function (simula aH)
  (let
    ((arrival-distribution (make-exponential 1)))
    (let loop ()
      (procesa (genera-cliente aH) aH)
      (wait-duration (sample arrival-distribution))
      (loop))))

;;=======================================================================
;;main de la simulacion
(set! resultados
      (let
        ((un-Hospital (instantiate Hospital :num-departamentos 5
                                            :num-clases-clientes 4)))
        (define (genera-departamento aH i max-n s-time)
          (vector-set! (Hospital-departamentos aH) i 
               (instantiate departamento
                   :service-distribution  s-time
                   :name (string-append "Dept" (number->String i)))))
        (genera-departamento un-Hospital 0 2 (make-exponential 3))
        (genera-departamento un-Hospital 1 2 (make-exponential 3))  
        (genera-departamento un-Hospital 2 2 (make-exponential 3))
        (genera-departamento un-Hospital 3 2 (make-exponential 3))
        (genera-departamento un-Hospital 4 2 (make-exponential 3))
        (init-simulation)
        (set! *random* (make-random))
        (trace-off *overlord*)  
        (create-event 100 stop )
        (simula un-Hospital)
        (start-simulation)
        (display-object un-Hospital)
        un-Hospital))
 
(display-object resultados)


