;;; Uncomment the following 2 lines if you are not using a compiled version

;(require 'monitors)
;(require 'distributions)

;;; Four machines tended by two workers workers have to produce one type of part
;;; To produce a part, a number of tasks have to be accomplished that require 
;;; coperation between a machine and a worker. Macine time, external time and
;;; internal time define the structure of a task. We assume standard workers. 
;;;
;;; We model the machines and workers as resources, by using facilities
(define numbering #f)
;;;
;;;
(define-class shop named-object
        ((= arrivals :inmutable)
         (= production  :inmutable)))
;----------------------- Tasks --------------------------------------------
(define-class task object
    (number machine-time external-time internal-time machine worker))
;--------------------------------------------------------------------------
(define-class machine facility ())
(define-class worker facility ())
(define-class part object 
              ((= n :initializer (lambda()(numbering)))
               (= lag :initializer (lambda()(make-exponential 5)))
               (= arrival-time :maybe-uninitialized)
               (= due-date :maybe-uninitialized)))

(define-generic (perform (aT task) aS))
(define-waitfor-method (perform (aT task) aS)
   (let ((wk  (task-worker aT))
         (mch (task-machine aT))
         (z (sample (task-external-time aT)))
         (x (sample (task-internal-time aT)))           
         (y (sample (task-machine-time  aT))))
     (if wk (enter wk))
     (if (> z 0) (wait-duration z ))
     (if mch (enter mch))
     (if (> x y)
       (begin (if mch (begin (wait-duration y)
                             (leave mch)))
              (wait-duration (- x y))
              (if wk (leave wk)))
       (begin (if wk (begin (if (> x 0) (wait-duration x))
                            (leave wk)))
              (if (> y x) (wait-duration (- y x)))
              (if mch (leave mch))))))


(define-generic (production-process (aP part) aS))
(define-async-method (production-process (aP part) aS)
   (let loop ((p (shop-production aS)))
     (if (pair? p)
       (begin
        (if (shop-trace aS)
          (displayLine (current-time) "perform "(task-number (car p))
                       " for " (part-n aP)))
        (perform (car p) aS)
        (if (shop-trace aS)
          (displayLine (current-time) "End  " (task-number (car p)) 
                       " for " (part-n aP)))
        (loop (cdr p))))))


(define-async-function (main-shop aS)
   (let loop ()
     (let ((x (instantiate part :arrival-time (current-time))))
       (initialize-field-value! x 
            (+ (current-time) (sample (part-lag x))) 'due-date)
       (production-process x aS)
       (wait-duration (sample (shop-arrivals aS)))
       (loop))))

;; ----------------------------  simulation ----------------------------------

(define mach1 (instantiate machine :max-number 1 :name 'mach1
                           :parking-place (instantiate stat-queue)))
(define mach2 (instantiate machine :max-number 1 :name 'mach2
                           :parking-place (instantiate stat-queue)))
(define mach3 (instantiate machine :max-number 1 :name 'mach3
                           :number (instantiate stime-var)
                           :parking-place (instantiate stat-queue)))
(define mach4 (instantiate machine :max-number 1 :name 'mach4
                           :parking-place (instantiate stat-queue)))

(define wk1 (instantiate worker :max-number 1 :name 'wk1))
(define wk2 (instantiate worker :max-number 1 :name 'wk2))

(define ashop
  (instantiate shop 
               :arrivals (make-exponential 130)
               :production (list
                             (make-task 1 10  5 2 mach1 wk1) 
                             (make-task 2 30 15 5 mach2 wk1)
                             (make-task 3  0 10 5  #f   wk1)
                             (make-task 4  0 10 3  #f   wk1)
                             (make-task 5  7  0 1 mach3 wk2)
                             (make-task 6 25 10 2 mach3 wk2)
                             (make-task 7 30  5 2 mach4 wk2))))

(begin
 (init-simulation)
 (set! numbering (make-counter))
 (set! *random* (make-random))
 (create-event 2000 stop )
 (main-shop ashop)
 (start-simulation)
 (display-object wk1)
 (display-object wk2)
 (display-object mach1)
 (display-object mach2)
 (display-object mach3)
 (display-object mach4))
