;; A stock point has inventory and assigns it to asking customers If a customer
;; does not find enough inventory waits until enough inventory is acumulated
;; only the first waiting customer is tested, when inventory is increased. If
;; the first does not pass nobody passes. If the first passes, the second
;; waiting customer is tested until either we run out of waiting customers or
;; run out of inventory.
;; This is not the general process of demand and supply, because demanding
;; customers are blocked until demand is met!!!!
;; This could be implemented with facilities, but I wanted to do test the 
;; conditions

;; A customer is an object with a time mark for the time of demand. 

(define-class customer named-object
     ((= quantity :maybe-uninitialized)
      (= time :initializer (lambda()0))))

;;----------------------------The stock point-------------------------
(define-class stock-point condition
    ((= stockt0 :initializer (lambda()0))
     (= stock :maybe-uninitialized)
     (= waiting-customers  :maybe-uninitialized)    
     (= service-delay :maybe-uninitialized)))

(define-handy-method (display-object (aS stock-point) . port)
    (display-object stock)
    (newline)
    (display-object waiting-customers)
    (newline)
    (display-object (stock-point-service-delay aS))
    (newline)
    (display-Histogram-for (stime-var-data (stock-point-service-delay aS))
                           10 (qualify-name aS 'Waiting)))

(define-method (initialize! (aSP stock-point))
    (call-next-method)
    (init-stats-gatherer! aSP
        (instantiate stime-var :name (qualify-name aSP 'stock))
        'stock)
    (initialize-field-value! aSP 
         (instantiate stat-queue :name (qualify-name aSP 'customer-in-wait)
                      :number 
                      (instantiate sio-var :name (qualify-name aSP 'IO-curves)))
         'waiting-customers)
    (initialize-field-value! aSP
         (instantiate stime-var :name (qualify-name aSP 'customer-wait))
         'service-delay)
    (take (stock-point-stock aSP) (make-point  (current-time)
                                               (stock-point-stockt0 aSP)))
    (reset aSP)
    aSP)

(define-generic (demand-from (aSP stock-point) xx))
(define-async-method (demand-from (aSP stock-point) xx)
   ;blocking operation!
    (if (or (< (var-value (stock-point-stock aSP)) (customer-quantity xx)) 
            (not (empty? (stock-point-waiting-customers aSP))))
      (begin
       (put! (stock-point-waiting-customers aSP) xx)
       (wait-for aSP) )))

(define-generic (supply-to (aSP stock-point) x))
(define-method (supply-to (aSP stock-point) x) 
;non-blocking operation
   (take (stock-point-stock aSP) 
         (make-point (current-time)
                     (+ (var-value (stock-point-stock aSP)) x)))
   (let loop ()
     (if (not (empty? (stock-point-waiting-customers aSP)))
       (let ((y (peek (stock-point-waiting-customers aSP))))
         (if (>= (var-value (stock-point-stock aSP)) (customer-quantity y))
           (let ((z (get! (stock-point-waiting-customers aSP))))
             (take (stock-point-stock aSP)
                   (make-point (current-time)
                               (- (var-value (stock-point-stock aSP))
                                  (customer-quantity z))))
             (take (stock-point-service-delay aSP)
                   (make-point (current-time)
                               (- (current-time)(customer-time z))))
             (if (stock-point-trace aSP)
               (displayLine  (get-name z)
                     " Received " (formatf (customer-quantity z) 6 2)
                     " left "
                     (formatf (var-value (stock-point-stock aSP))  6 2)
                     " delay "
                     (formatf (- (current-time)(customer-time z)) 10 2)))
             (signal aSP)
             (loop)))))))

;; The processes--------------------------------------------------------
(define-generic (proveedor (aS stock-point) dist-time dist-ammount))
(define-async-method (proveedor (aS stock-point)  dist-time dist-ammount)
 (let loop ((ammount (sample dist-ammount)))
      (wait-duration (sample dist-time))
       (if (stock-point-trace aS)
         (displayLine " T= " (Current-time) 
                      " Served" (formatf ammount 6 2)
                      " stock " 
                      (formatf (var-value (stock-point-stock aS)) 6 2)))
      (supply-to aS ammount)
      (loop (sample dist-ammount))))

(define-generic (pedidor (aS stock-point) dist-time dist-ammount))
(define-async-method (pedidor (aS stock-point) dist-time dist-ammount)
  (let ((count (make-counter)))
    (let loop ((xx (instantiate customer :quantity (sample dist-ammount)
                                :name (qualify-name aS (count)))))
      (wait-duration (sample dist-time))
      (set-customer-time! xx (current-time))
       (if (stock-point-trace aS)
         (displayLine (get-name xx) 
                      " T=" (current-time) " pedidas "
                      (formatf (customer-quantity xx) 6 2)
                      " stock "
                      (formatf (var-value (stock-point-stock aS)) 6 2)))
      (demand-from aS xx)
      (loop (instantiate customer :quantity (sample dist-ammount)
                         :name (qualify-name aS (count)))))))

;;---------------The simulation-----------------------------------------
(set! *random* (make-random))
(init-simulation)
;The next sentence is here because uses current-time, initialized by 
;init-simulation!
(define chan (instantiate stock-point :name 'Service-Center :stockt0 0)) 
(create-event 1000 stop)
 ;if you want to see what is going on, uncomment the next sentence
;(trace-on chan) 
(proveedor chan (make-exponential 10)(make-exponential 10))
(pedidor chan (make-exponential 5)(make-exponential 5))
(start-simulation)
(display-object chan)
                                    
                                    
